from djitellopy import Tello
import numpy as np
import cv2

def detect_arrow_directions(frame):
    """ Detects the arrow direction """
    processed_frame = frame.copy()
    direction = None

    gray_frame = cv2.cvtColor(processed_frame, cv2.COLOR_BGR2GRAY) # Converts the colored frame into a gray frame.
    _, thresh = cv2.threshold(gray_frame, 150, 255, cv2.THRESH_BINARY_INV) # Converts the frame into white-black balanced (the color between 0-150 into white and 150-255 into black).

    kernel = np.ones((5, 5), np.uint8)
    thresh_0 = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel, iterations = 2) # cv2.morphologyEx() fills these gaps, making objects more continuous and well-defined.
    # Also, Morphological closing (cv2.MORPH_CLOSE) applies:
    # - Dilation (expanding white regions).
    # - Erosion (shrinking white regions).
    # So, we can close the noisy area in this case.

    contours, _ = cv2.findContours(thresh_0, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    # cv2.RETR_EXTERNAL: Only outermost contours are detected (ignoring nested ones).
    # The method cv2.CHAIN_APPROX_SIMPLE compresses redundant points, reducing memory usage.
    # So contours, gives us the point's coordinates between the detected colour difference which are between black and white which also it works for detecting shapes and etc.
    if not contours:
        return processed_frame, None
    
    # The code structure given below aims to reduce the noises (The contour sizes under 800)
    largest_contour = max(contours, key = cv2.contourArea)
    area = cv2.contourArea(largest_contour)
    if area < 800: # Just like mentioned above, the aim is to reduce the noises detected.
        return processed_frame, None
    
    hull_points = cv2.convexHull(largest_contour, returnPoints = True) # This detects the point's coordinates of the largest contour's convex.
    max_distance, farthest_point = 0, None
    x, y, w, h = cv2.boundingRect(largest_contour)

    for point in hull_points:
        pt = tuple(point[0])
        dx, dy = pt[0] - ((x + w)//2), pt[1] - ((y + h)//2)
        distance = dx**2 + dy**2
        if distance > max_distance:
            max_distance, farthest_point = distance, pt
    
    if not farthest_point:
        return processed_frame, None
    
    # We now, take the farthest points to begin using it.
    dx_farthest_p = farthest_point[0] - (x + w)//2
    # dy_farthest_p = (y + h)//2 - farthest_point[1]
    dy_farthest_p = farthest_point[1] - (y + h)//2

    angle = np.arctan2(dy_farthest_p, dx_farthest_p) * 180 / np.pi
    if angle < 0:
        angle += 360

    if 45 < angle <= 135:
        direction = "UP"
    elif 135 < angle <= 225:
        direction = "LEFT"
    elif 225 < angle <= 315:
        direction = "DOWN"
    else:
        direction = "RIGHT"
    
    cv2.drawContours(processed_frame, [largest_contour], -1, (0, 255, 0), 2) # The output contour will be represented in green colour.
    cv2.circle(processed_frame, farthest_point, 5, (0, 0, 255), -1) # The circle will be red coloured.
    cv2.putText(processed_frame, direction, (x, y - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 0), 2) # This writes the direction into the processed frame
    # within font size .7 and FONT_HERSHEY_SIMPLEX with green colour.

    return processed_frame, direction

def video_stream(drone):
    """ Handles the video stream """
    recording = False
    video_writer = None
    frame_width = 960
    frame_height = 720
    fps = 20

    try:
        while True:
            file_number = 0
            video_number = 0

            drone_frame = drone.get_frame_read().frame
            if drone_frame is None or drone_frame.size == 0:
                continue

            processed_frame, direction  = detect_arrow_directions(frame = drone_frame)

            cv2.imshow("Drone View", processed_frame)

            key = cv2.waitKey(1) & 0xFF

            if key == ord('q'):
                print("Closing frame...")
                break

            elif key == ord('c'):
                file_name = f"{file_number}_drone_photo.jpg"
                file_number += 1
                cv2.imwrite(file_name, processed_frame)
                print(f"'{file_name}' photo captured.")

            elif key == ord('r'):
                if not recording:
                    print("Recording Started")
                    video_writer = cv2.VideoWriter(f"drone_video_{video_number}.avi", cv2.VideoWriter_fourcc(*"XVID"), fps, (frame_width, frame_height))
                    recording = True
                    video_number += 1

                else:
                    print("Recording Stopped")
                    recording = False
                    video_writer.release()
                    video_writer = None

            if recording and video_writer is not None:
                video_writer.write(processed_frame)

    except Exception as e:
        print(f"Unexpected error appeared: {e}")

    finally:
        if recording and video_writer is not None:
            video_writer.release()

        cv2.destroyAllWindows()

def main():
    """ This displays the main function """
    tello_drone = Tello()

    try:
        tello_drone.connect()
        tello_drone.streamon()

        print(f"Drone Battery: {tello_drone.get_battery()}%")

        video_stream(drone = tello_drone)

    except KeyboardInterrupt as k:
        print(f"Keyboard interrupt: {k}")

    finally:
        tello_drone.streamoff()
        tello_drone.end()
    
if __name__ == "__main__":
    main()